// Copyright  1999 Branko Dimitrijevic
// bepp@bepp.8m.com
// http://bepp.8m.com

unit DBSortComboBox;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, db, BDB_Shared;

type

  TDBSortEvent = procedure(Sender: TObject; DataSet: TDataSet; const SortFieldNames: string) of object;

  TDBSortComboBox = class(TComboBox)
  private
    FAutoUpdate: boolean;
    FDataLink: TSortDataLink;
    FReopenAfterError: boolean;
    FStrictServerIndexing: boolean;
    FAfterSortChange: TDataSetNotifyEvent;
    FOnSort: TDBSortEvent;
    procedure SetDataSource(const Value: TDataSource);
    function GetDataSource: TDataSource;

    function IsValidTableDataSource: boolean;

    procedure DoAfterSortChange(Sender: TObject; const OldFields,NewFields: string; var Done: boolean);
    procedure SetAutoUpdate(const Value: boolean);

  protected
    procedure Click; override;
    procedure DoOnSort(DataSet: TDataSet; const SortFieldNames: string);
    procedure DropDown; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Sort; dynamic;
    //procedure Loaded; override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Refresh;

  published
    // Properties.
    property AutoUpdate: boolean read FAutoUpdate write SetAutoUpdate;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property ReopenAfterError: boolean read FReopenAfterError write FReopenAfterError;
    property StrictServerIndexing: boolean read FStrictServerIndexing write FStrictServerIndexing;

    // Events.
    property AfterSortChange: TDataSetNotifyEvent read FAfterSortChange write FAfterSortChange;
    property OnSort: TDBSortEvent read FOnSort write FOnSort;
  end;

procedure Register;

implementation

{$R *.dcr}

uses
  dbtables
  {$IFDEF VER130}
  ,IBTable
  {$ENDIF}
  ;

procedure Register;
begin
  RegisterComponents('BDB', [TDBSortComboBox]);
end;

{ TDBSortComboBox }

procedure TDBSortComboBox.Click;
begin
  inherited;
  Sort
end;

constructor TDBSortComboBox.Create(AOwner: TComponent);
begin
  inherited;
  FAutoUpdate := true;
  FDataLink := TSortDataLink.Create;
  FDataLink.AfterSortChange := DoAfterSortChange;
  FReopenAfterError := true
end;

destructor TDBSortComboBox.Destroy;
begin
  FDataLink.Free;
  inherited;
end;

procedure TDBSortComboBox.DoAfterSortChange(Sender: TObject; const OldFields,NewFields: string; var Done: boolean);

  function LeftmostField(Fields: string): string;
  var
    i,len: integer;
  begin
    len := Length(Fields);
    for i := 1 to len do
      if Fields[i] = ';' then
      begin
        result := Copy(Fields,1,i-1);
        exit
      end;

    result := Fields
  end;

var
  i: integer;
  LeftmostFld: string;
  ItemIdx: integer; // To avoid
begin
  if AutoUpdate and IsValidTableDataSource then
  begin
    //Text := ExtractSortFieldNames((Sender as TDataLink).DataSet);
    ItemIdx := -1;
    LeftmostFld := LeftmostField(NewFields);
    for i := 0 to Items.Count-1 do
    begin
      if Items[i] = NewFields then
      begin
        ItemIdx := i;
        break
      end
      else if Items[i] = LeftmostFld then
        ItemIdx := i;
    end; // for i := 0 to Items.Count-1

    if ItemIdx >= 0 then
      ItemIndex := ItemIdx;

    Text := NewFields;
    if Assigned(FAfterSortChange) then
      FAfterSortChange((Sender as TDataLink).DataSet)
  end
end;

procedure TDBSortComboBox.DoOnSort(DataSet: TDataSet;
  const SortFieldNames: string);
begin
  if Assigned(FOnSort) then
    FOnSort(self,DataSet,SortFieldNames)
end;

procedure TDBSortComboBox.DropDown;
var
  i: integer;
  Tmp: string;
  TmpLen,MasterLen,MinLen: integer;
begin

  if Assigned(DataSource) and Assigned(DataSource.DataSet) then
  begin

    Items.Clear;

    if DataSource.DataSet is TTable then
    begin

      with DataSource.DataSet as TTable do
      begin

        if IsSequenced or StrictServerIndexing then
        begin
          IndexDefs.Update;

          for i := 0 to IndexDefs.Count-1 do
          begin
            Tmp := IndexDefs[i].Fields;
            if Assigned(MasterSource) and not (MasterFields = '') then
            begin

              TmpLen := Length(Tmp);
              MasterLen := Length(MasterFields);
              if TmpLen <= MasterLen then
                MinLen := TmpLen
              else
                MinLen := MasterLen;

              // New index begins with MasterFields?
              if
                (Copy(MasterFields,1,MinLen) = Copy(Tmp,1,MinLen))
                and (
                  (MasterLen = TmpLen)
                  or (
                    (MasterLen < TmpLen)
                    and (Tmp[MasterLen+1] = ';')
                  )
                )
              then
                Items.Append(Tmp)
            end
            else
              Items.Append(Tmp);

          end // for i := 0 to IndexDefs.Count-1

        end // if IsSequenced or StrictServerIndexing

        else
          for i := 0 to FieldCount-1 do
          begin
            if (Fields[i].FieldKind = fkData) and not (Fields[i].IsBlob) then
            begin
              Tmp := Fields[i].FieldName;
              if Assigned(MasterSource) and not (MasterFields = '') and not (MasterFields = Tmp) then
                Items.Append(MasterFields + ';' + Tmp)
              else
                Items.Append(Tmp)
            end
          end

      end // with DataSource.DataSet as TTable

    end // if DataSource.DataSet is TTable

    else
      with DataSource.DataSet do
        for i := 0 to FieldCount-1 do
          if (Fields[i].FieldKind = fkData) and not (Fields[i].IsBlob) then
            Items.Append(Fields[i].FieldName);

  end; // if IsValidDataSource

  inherited;

end;

function TDBSortComboBox.GetDataSource: TDataSource;
begin
  result := FDataLink.DataSource
end;

function TDBSortComboBox.IsValidTableDataSource: boolean;
begin
  result :=
    Assigned(DataSource)
    and Assigned(DataSource.DataSet)
    and
      {$IFDEF VER130}
        (
      {$ENDIF}
          (DataSource.DataSet is TTable)
      {$IFDEF VER130}
          or (DataSource.DataSet is TIBTable)
        )
      {$ENDIF}
end;

procedure TDBSortComboBox.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited;
  if Key = VK_RETURN then
    Sort
end;

procedure TDBSortComboBox.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) and (AComponent = DataSource) then
    DataSource := nil
end;

(*
procedure TDBSortComboBox.Loaded;
begin
  inherited;
  if IsValidDataSource then
    Text := ExtractSortFieldNames(DataSource.DataSet)
end;
*)

procedure TDBSortComboBox.Refresh;
begin
  inherited;
  if IsValidTableDataSource then
    Text := ExtractSortFieldNames(DataSource.DataSet)
end;

procedure TDBSortComboBox.SetAutoUpdate(const Value: boolean);
begin
  FAutoUpdate := Value;
  if Value then
    Refresh;
end;

procedure TDBSortComboBox.SetDataSource(const Value: TDataSource);
begin
  FDataLink.DataSource := Value;
  Refresh
end;

procedure TDBSortComboBox.Sort;
var
  len: integer;
  OldIndexFieldNames: string;
begin

  if IsValidTableDataSource then
  begin

    try

      if DataSource.DataSet is TTable then

        with TTable(DataSource.DataSet) do
        begin

          len := Length(MasterFields);
          if
            Assigned(MasterSource)
            and (len > 0)
            and not (
              (Copy(Text,1,len) = MasterFields)
              and (
                (Length(Text) = len)
                or (Text[len+1] = ';')
              )
            )
          then
          begin
            if Text = '' then
              Text := MasterFields
            else
              Text := MasterFields + ';' + Text;
          end;

          try
            OldIndexFieldNames := IndexFieldNames;
            IndexFieldNames := Text;
          except
            IndexFieldNames := OldIndexFieldNames;
            raise
          end

        end // with DataSource.DataSet as TTable


      {$IFDEF VER130}
      else if DataSource.DataSet is TIBTable then

        with TIBTable(DataSource.DataSet) do
        begin

          len := Length(MasterFields);
          if
            Assigned(MasterSource)
            and (len > 0)
            and not (
              (Copy(Text,1,len) = MasterFields)
              and (
                (Length(Text) = len)
                or (Text[len+1] = ';')
              )
            )
          then
          begin
            if Text = '' then
              Text := MasterFields
            else
              Text := MasterFields + ';' + Text;
          end;

          try
            OldIndexFieldNames := IndexFieldNames;
            IndexFieldNames := Text;
          except
            IndexFieldNames := OldIndexFieldNames;
            raise
          end

        end // with TIBTable(DataSource.DataSet)

      {$ENDIF}
      else
        raise Exception.Create('TDBSortComboBox.Sort() Internal error: unknown DataSet type.')

    except
      on X: Exception do
      begin
        if
          FReopenAfterError
          and Assigned(DataSource) // Just in case.
          and Assigned(DataSource.DataSet) // Just in case.
          and not DataSource.DataSet.Active
        then
          DataSource.DataSet.Open;
        raise
      end
    end

  end; // if IsValidTableDataSource

  DoOnSort(DataSource.DataSet,Text)

end;



end.
